# INSERT LICENSE
# This script contains utility functions
# Importing dependencies

from typing import Optional, Union, Any

import navis
import networkx as nx
import numpy as np
import pymaid

# Calculating cable length between nodes - this function does not exist for navis neurons


def check_valid_neuron_input(x: navis.TreeNeuron):
    """
    Takes an object and checks whether it is a navis TreeNeuron object

    Parameters
    --------
    x: a python object: hopefully a navis.TreeNeuron object

    Returns
    --------
    x: returns x if it is a navis.TreeNeuron object, otherwise will raise an assertion error

    """
    assert isinstance(
        x, (navis.TreeNeuron, navis.NeuronList)
    ), f"Need to pass a Navis Tree Neuron type. You have passed: {type(x)}"

    if isinstance(x, navis.NeuronList):
        assert len(x) < 1, "Need to pass a SINGLE Neuron"
        x = x[0]
    return x


def calc_cable(
    x: Union[navis.TreeNeuron, navis.NeuronList], return_skdata: bool = False
):

    """
    # Calculate cable length between nodes - this function does not exist for navis neurons
    :param x:
    :param return_skdata:
    :return:
    """

    x = check_valid_neuron_input(x)

    nodes = x.nodes[~x.nodes.parent_id.isnull()]
    tn_coords = nodes[["x", "y", "z"]].values

    this_tn = nodes.set_index("node_id")
    parent_coords = this_tn.reindex(index=nodes.parent_id.values,
                                    columns=["x", "y", "z"]).values

    w = np.sqrt(np.sum((tn_coords - parent_coords) ** 2, axis=1))

    if return_skdata:
        nodes.loc[~nodes.parent_id.isnull(), "parent_dist"] = w / 1000
        x.nodes = nodes
        return x

    return np.sum(w[np.logical_not(np.isnan(w))]) / 1000


def check_valid_pymaid_input(x: Any) -> Optional[pymaid.core.CatmaidNeuron]:

    """
    Takes an object and checks whether it is a navis TreeNeuron object

    Paramters
    --------
    x: a python object: hopefully a navis.TreeNeuron object

    Returns
    --------
    x: returns x if it is a navis.TreeNeuron object, otherwise will raise an assertion error

    """
    assert isinstance(
        x, (pymaid.core.CatmaidNeuron, pymaid.core.CatmaidNeuronList)
    ), f"Need to pass a Navis Tree Neuron type. You have passed: {x}"

    if isinstance(x, pymaid.core.CatmaidNeuron):
        assert len(x) < 1, "Need to pass a SINGLE Neuron"
        x = x[0]
    return x


def pymaid_topological_sort(
    x: Union[pymaid.core.CatmaidNeuron, pymaid.core.CatmaidNeuronList],
    return_object: str = "list",
):

    """
    Takes pymaid/CatmaidNeuron and topologically sorts the nodes

    Paramters
    ---------
    x:                a pymaid/Catmaid neuron object
    return_skdata:    bool
                    whether to return a list of node_ids topologically sorted
                    or to return a dict where the keys are treenodes and the values
                    are ranking in the topological sort

    Returns
    --------
    x: list or dict

    Examples
    --------
    """

    x = check_valid_pymaid_input(x)

    g = nx.DiGraph()
    g.add_node(x.nodes[x.nodes.type == "root"].treenode_id.values[0])
    g.add_nodes_from(
        [
            i
            for i in x.nodes.node_id.tolist()
            if i != x.nodes[x.nodes.type == "root"].treenode_id.values[0]
        ]
    )

    for e in x.nodes[["treenode_id", "parent_id"]].values:

        if e[1] is not None:

            g.add_edge(e[1], e[0])

        else:

            continue

    topological_sort = [i for i in nx.topological_sort(g)]

    if return_object == "list":
        return topological_sort
    elif return_object == "dict":
        topological_sort = dict(zip(topological_sort, range(0, len(topological_sort))))
        return topological_sort


def pymaid_to_navis(x: Union[pymaid.core.CatmaidNeuron, pymaid.core.CatmaidNeuronList]):

    """
    Takes pymaid/CatmaidNeuron and topologically sorts the nodes

    Paramters
    ---------
    x:                a pymaid/Catmaid neuron object

    return_skdata:    bool
                    whether to return a list of node_ids topologically sorted
                    or to return a dict where the keys are treenodes and the values
                    are ranking in the topological sort

    Returns
    --------
    x: list or dict

    Examples
    --------
    """

    x = check_valid_pymaid_input(x)

    x.nodes["rank"] = x.nodes.treenode_id.map(
        pymaid_topological_sort(x, return_object="dict")
    )
    x.nodes.sort_values(by=["rank"], ascending=True, inplace=True)

    # Getting the topological sort of the pymaid neuron
    x_graph = x.graph
    navis_neuron = navis.TreeNeuron(x_graph)

    # Populating the xyz columns of nodes

    for i, j in enumerate(x.nodes[["x", "y", "z"]].values):

        navis_neuron.nodes.loc[i, "x"] = j[0]
        navis_neuron.nodes.loc[i, "y"] = j[1]
        navis_neuron.nodes.loc[i, "z"] = j[2]

    # adding type column
    navis_neuron.nodes["type"] = x.nodes.type.copy()

    # adding connectors
    navis_neuron.connectors = x.connectors.copy()
    navis_neuron.connectors = [
        "pre" if i == 0 else "post" for i in navis_neuron.connectors.type
    ]

    # adding soma & name

    navis_neuron.soma = x.soma
    navis_neuron.name = x.neuron_name

    return navis_neuron



import navis
from tqdm import tqdm
from random import randrange
import cloudvolume
import time
from typing import Union, Tuple
import numpy as np
from itertools import chain


def random_sample_from_volume(
    volume: navis.Volume,
    supervoxels: cloudvolume.frontends.precomputed.CloudVolumePrecomputed,
    segmentation: cloudvolume.frontends.graphene.CloudVolumeGraphene,
    amount_to_query: Union[float, int] = 10,
    mip_level: Tuple[int, int, int] = (64, 64, 40),
    n_points: int = int(1e6),
    bbox_dim: int = 20,
    disable_rand_point_progress: bool = False
):

    """
    Takes a FAFB14 volume and generates random points within it.
    These points are then used to create a series of small bounding boxes.
    The supervoxels within these bounding boxes are then queried.
    These supervoxel ids are then mapped to root ids.

    Parameters
    ----------

    volume: a navis.Volume object retrieved either through FAFB14 or Janelia Hemibrain.
                If the latter, then you need to transform the volume into flywire
                space using a bridging transformation before using this function

    mip_level: the mip resolution level.
                Available resolutions are (16, 16, 40), (32, 32, 40), (64, 64, 40).
                I recommend the lowest voxel resolution (64, 64, 40).

    n_points: the number of randomly generated points to create.
                Note that not all of these will be in the volume, although those that are are kept.

    bbox_dim: dimension of the query cube - small values (<100) are encouraged.

    supervoxels: A supervoxel cloudvolume.Volume object
                specifying the supervoxel instance you are querying.

    e.g. cloudvolume.CloudVolume("precomputed://https://s3-hpcrc.rc.princeton.edu/fafbv14-ws/ws_190410_FAFB_v02_ws_size_threshold_200",
                                mip=mip_level)

    segmentation: A segmentation cloudvolume.Volume object specifying the segmentation instance you are querying.
                e.g. cloudvolume.CloudVolume("graphene://https://prodv1.flywire-daf.com/segmentation/table/fly_v31")

    amount_to_query: Either a float or an int.
                If a float, this is a percentage of the total number of randomly generated points that are in the volume you want to query
                If an int, the first n randomly generated points that are in the volume will be used to query.

    Returns
    ----------

    root_ids: an array of root ids that were within the randomly generated query boxes.


    Example
    ----------
    import pymaid
    import cloudvolume
    import numpy as np
    import navis
    from random import randrange
    import chain
    import time

    MB_R = pymaid.get_volume('MB_whole_R')
​
    mip_level = (64, 64, 40)
​
    cv_svs = cloudvolume.CloudVolume("precomputed://https://s3-hpcrc.rc.princeton.edu/fafbv14-ws/ws_190410_FAFB_v02_ws_size_threshold_200", mip=mip_level)

    cv_seg = cloudvolume.CloudVolume("graphene://https://prodv1.flywire-daf.com/segmentation/table/fly_v31")

    root_ids = random_sample_from_volume(volume = MB_R, mip_level = mip_level,
                                 n_points = int(1e6), bbox_dim = 20,
                                 supervoxels=cv_svs, segmentation=cv_seg)

    """
    start_time = time.time()

    assert isinstance(volume, navis.Volume), f'You need to pass a navis.Volume object. You passed {type(volume)}.'
    assert isinstance(supervoxels, cloudvolume.frontends.precomputed.CloudVolumePrecomputed), f'You need to pass a cloudvolume.frontends.precomputed.CloudVolumePrecomputed object. You passed {type(supervoxels)}.'
    assert isinstance(segmentation, cloudvolume.frontends.graphene.CloudVolumeGraphene), f'You need to pass a cloudvolume.frontends.graphene.CloudVolumeGraphene object. You passed {type(segmentation)}.'

    if type(n_points) != int:

        n_points = int(n_points)

    volume.vertices /= mip_level

    vertices_array = np.array(volume.vertices)
    vertices_array = vertices_array.astype(int)

    # generating random points
    n_point_string = format(n_points, ',d')
    print(f'Generating {n_point_string} random points... \n')

    # finding the min and max values of each xyz dimension
    x_min = min(vertices_array[:, 0])
    x_max = max(vertices_array[:, 0])

    y_min = min(vertices_array[:, 1])
    y_max = max(vertices_array[:, 1])

    z_min = min(vertices_array[:, 2])
    z_max = max(vertices_array[:, 2])

    # randomly generating integers inbetween these max and min values
    rand_x = [randrange(x_min, x_max) for i in tqdm(range(n_points), disable = disable_rand_point_progress)]
    rand_y = [randrange(y_min, y_max) for i in tqdm(range(n_points), disable = disable_rand_point_progress)]
    rand_z = [randrange(z_min, z_max) for i in tqdm(range(n_points), disable = disable_rand_point_progress)]

    xyz_arr = np.array([[rand_x, rand_y, rand_z]])
    xyz_arr = xyz_arr.T
    xyz_arr = xyz_arr.reshape(n_points, 3)

    # How many randomly generated points are in the volume?
    in_vol = navis.in_volume(xyz_arr, volume)
    print(f"""Of {n_point_string} random points generated, {xyz_arr[in_vol].shape[0] / n_points * 1e2 :.2f}% of them are in the volume.""")
    print(f"""This equals {xyz_arr[in_vol].shape[0]} points. \n""")
    xyz_in_vol = xyz_arr[in_vol]

    # generating query cubes
    xyz_start = xyz_in_vol
    xyz_end = xyz_in_vol + bbox_dim

    # querying flywire
    print('Querying flywire... \n')

    supervoxel_ids = []

    if isinstance(amount_to_query, float):

        assert amount_to_query <= 1.0, '''If using percentages of the total number of randomly generated points,
                                        you cannot use more than 100% of the points generated.
                                        Did you intend to use an integer?'''

        print(f'You are passing a float, so using {amount_to_query * 1e2}% of the total number ({len(xyz_start)}) of randomly points generated. \n')

        if amount_to_query == 1.0:

            print('You are using 100% of the randomly generated points - this can take a long time to complete.')

        n_query = int(len(xyz_start) * amount_to_query)

        print(f'{amount_to_query * 1e2}% = {n_query} points')

        xyz_start = xyz_start[:n_query]
        xyz_end = xyz_end[:n_query]

        print(f'Coverage: {((bbox_dim ** 3) * n_query / volume.volume) * 1e2 :.2g} % of the total {volume.name} volume is covered with {n_query} bounding boxes of {bbox_dim} cubic dimensions. \n')

        for i in range(n_query):

            q = supervoxels[xyz_start[i][0]: xyz_end[i][0],
                            xyz_start[i][1]: xyz_end[i][1],
                            xyz_start[i][2]: xyz_end[i][2]]

            supervoxel_ids.append(q)

    elif isinstance(amount_to_query, int):

        assert amount_to_query <= len(xyz_start), f'''You cannot use the first {amount_to_query} randomly generated
                                                    points when you have only {len(xyz_start)} exist in the volume.
                                                    Increase the number of randomly generated points in n_points.'''

        print(f'You are passing an integer, so the first {amount_to_query} randomly generated points will be used to query. \n')

        xyz_start = xyz_start[:amount_to_query]
        xyz_end = xyz_end[:amount_to_query]

        print(f'Coverage: {((bbox_dim ** 3) * amount_to_query / volume.volume) * 1e2 :.2g} % of the total {volume.name} volume is covered with {amount_to_query} bounding boxes of {bbox_dim} cubic dimensions. \n')

        for i in range(amount_to_query):

            q = supervoxels[xyz_start[i][0]: xyz_end[i][0],
                            xyz_start[i][1]: xyz_end[i][1],
                            xyz_start[i][2]: xyz_end[i][2]]

            supervoxel_ids.append(q)

    sv_ids_unique = [np.unique(i) for i in supervoxel_ids]
    sv_ids_unique = np.unique(list(chain.from_iterable(sv_ids_unique)))
    print('Fetching root ids... \n')
    root_ids = segmentation.get_roots(sv_ids_unique)

    # Removing zeros
    root_ids = root_ids[~(root_ids == 0)]

    root_ids = np.unique(root_ids)

    print(f'Random sampling complete: {len(root_ids)} unique root ids found \n')
    print(f'This function took {(time.time() - start_time) / 60 :.2f} minutes to complete')
    return(root_ids)



# INSERT LICENSE
from collections import Counter
from typing import List, Optional

import navis
import navis.interfaces.neuprint as nvneu
import pandas as pd

# Create an adjacency matrix from synapse connections (neuprint)


def adjx_from_syn_conn(
    x: List[int],
    presyn_postsyn: str = "pre",
    roi: Optional = None,
    ct: tuple = (0.0, 0.0),
    rename_index: bool = False,
):

    """
    Creates an adjacency matrix from synapse connections

    Parameters
    ----------
    presyn_postsyn:
    x :                 list
                        a list of bodyIds of which you want to find their synaptic (pre/post) connections

    presyn_postsyn :    str
                        a string of either 'pre' or 'post'
                        If 'pre', the function will search for the
                        presynaptic connections / downstream neurons of x

                        If 'post', the function will search for the
                        postsynaptic connections / upstream neurons of x

    roi :               navis.Volume
                        A region of interest within which you are filtering the connections for


    ct :                tuple
                        Confidence threshold tuple containing the confidence value to filter above
                        The first value is presynaptic confidence and the second value postsynaptic confidence
                        e.g. (0.9, 0.8) will filter for connections where the presynaptic confidence > 0.9
                        and the postsynaptic confidence > 0.8

    rename_index :      bool
                        Whether to rename the index using the type of the connected neuron


    Returns
    -------
    df :                a DataFrame where x are the columns, the connection type (pre/post) are the rows
                        and the values the number of connections

    partner_type_dict : a dictionary where the keys are bodyIds of the
                        upstream/downstream neurons and the values are their types

    Examples
    --------
    """

    if presyn_postsyn == "pre":

        con = nvneu.fetch_synapse_connections(source_criteria=x)

        if roi:

            tt = navis.in_volume(con[["x_pre", "y_pre", "z_pre"]].values, roi)
            con = con[tt].copy()

        if ct[0] or ct[1] > 0.0:

            con = con[(con.confidence_pre > ct[0]) & (con.confidence_post > ct[1])].copy()

        neurons = con.bodyId_post.unique()
        n, _ = nvneu.fetch_neurons(neurons)
        partner_type_dict = dict(zip(n.bodyId.tolist(), n.type.tolist()))

        count = Counter(con.bodyId_post)
        count = count.most_common()
        count_dict = dict(count)

        df = pd.DataFrame(columns=[x], index=[i for i, j in count], data=[count_dict[i] for i, j in count])

    elif presyn_postsyn == "post":

        con = nvneu.fetch_synapse_connections(target_criteria=x)

        if roi:

            tt = navis.in_volume(con[["x_post", "y_post", "z_post"]].values, roi)
            con = con[tt].copy()

        if ct[0] or ct[1] > 0.0:

            con = con[(con.confidence_pre > ct[0]) & (con.confidence_post > ct[1])].copy()

        neurons = con.bodyId_pre.unique()
        n, _ = nvneu.fetch_neurons(neurons)
        partner_type_dict = dict(zip(n.bodyId.tolist(), n.type.tolist()))

        count = Counter(con.bodyId_pre)
        count = count.most_common()
        count_dict = dict(count)

        df = pd.DataFrame(index=[i for i, j in count], columns=[x], data=[count_dict[i] for i, j in count])
        df = df.T.copy()

    # df = pd.DataFrame(
    #     columns=[x], index=[i for i, j in count], data=[count_dict[i] for i, j in count])

    if rename_index:

        df.index = [partner_type_dict[i] for i in df.index]

    return (df, partner_type_dict)
